/*
Copyright (C) Christopher Fogelberg, Mengjie Zhang 2005

This file is part of VUWLGP.

VUWLGP is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

VUWLGP is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VUWLGP; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
  
#include <ctime>
#include <iostream>

#include "Config.h"
#include "GenerationalTournamentPopulation.h"

#include "InstructionArgumentConstant.h"
#include "InstructionArgumentFeature.h"
#include "InstructionArgumentRegister.h"

#include "InstructionOperationPlus.h"
#include "InstructionOperationMinus.h"
#include "InstructionOperationMult.h"
#include "InstructionOperationDiv.h"
#include "InstructionOperationIflt.h"

#include "MultiClassProgram.h"

using namespace VUWLGP; // else the lines of code get stupidly and clumsily long

// Initialise the static members which specify the proportion selected through elitism and
// the size of the selection tournaments for other evolutionary operators:
double GenerationalTournamentPopulation<MultiClassProgram<double>, double>::proportionElitism 
  = 0.1;
unsigned int GenerationalTournamentPopulation<MultiClassProgram<double>, double>::tournamentSize
  = 4;

int main(int argc, char** argv) {
  // Building the relevant config object - specify it all, even though not all of it is needed:
  Config<double>* c;
  if(argc == 1) {
    c = new Config<double>();
  }
  else {
    c = new Config<double>(argc, argv, true);
  }

  c->numRegisters = 4;
  c->numFeatures = 3;
  c->epsilon = 0;

  c->argumentGenerators->AddElement(InstructionArgumentConstant<double>::Generate);
  c->argumentGenerators->AddElement(InstructionArgumentFeature<double>::Generate);
  c->argumentGenerators->AddElement(InstructionArgumentRegister<double>::Generate);

  c->instructionOperations->AddElement(InstructionOperationPlus<double>::Generate);
  c->instructionOperations->AddElement(InstructionOperationMinus<double>::Generate);
  c->instructionOperations->AddElement(InstructionOperationMult<double>::Generate);
  c->instructionOperations->AddElement(InstructionOperationDiv<double>::Generate);
  c->instructionOperations->AddElement(InstructionOperationIflt<double>::Generate);

  c->initialMinLength = 6;
  c->initialMaxLength = 10;
  c->maxLength = 10;
  c->populationSize = 500;

  c->maxGenerations = 50;
  
  c->popLogInterval = 1;
  c->popLogFilePath = "population";
  c->statsLogFilePath = "log";

  c->runLogFilePath = "results.csv";

  // Initialise the RNG:
  if(c->seedSpecified) {
    Rand::Init(c->randSeed);
  }
  else {
    Rand::Init();
  }

  // Building the population object, randomising the fitness of it all:
  IPopulation<MultiClassProgram<double>, double>* pop = 
    new GenerationalTournamentPopulation<MultiClassProgram<double>, double>(c);

  // Build the fitness environment we will train and another we will test on:
  FitnessEnvironment<double> train(c);
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   6.7 -0.2 1.3"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   1.2  0.9 0.6"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   4.8  3.9 0.1"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second -0.2  6.7 1.3"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second  0.9  1.2 0.6"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second  3.9  4.8 0.1"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third   1.3 -0.2 6.7"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third   0.6  0.9 1.2"));
  train.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third   0.1  3.9 4.8"));

  FitnessEnvironment<double> test(c);
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   0.02 -0.3   0"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   0.99  0.4   0.7"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 0 first   9.76  0     3.2"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second  0     0.02 -0.3"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second  0.7   0.99  0.4"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 1 second  3.2   9.76  0"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third  -0.3   0     0.02"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third   0.4   0.7   0.99"));
  test.AddCase(new MultiClassFitnessCase<double>("fileName x y 2 third   0     3.2   9.76"));

  // Carry out the evolution:
  clock();
  unsigned int generationsUsedOrNoSolution = pop->Evolve(train);
  int time = clock();

  if(generationsUsedOrNoSolution <= c->maxGenerations) {
    std::cout << "Solution found.\nSolution is:" << std::endl;
  }
  else {
    std::cout << "No solution found.\nBest program:" << std::endl;
  }

  IProgram<double>* best = pop->GetFittestProgram();
  std::cout << pop->GetFittestProgram()->ToString(true, true) << std::endl;

  // Now calculate the statistics that aren't already calculated:
  std::cout << "Writing run statistics to: " << c->runLogFilePath << std::endl;

  // Time is already set, generations is set as it is because if no solution is found it comes
  // back set to maxGen + 1 (to distinguish the return value from "solution found in last gen").
  // Test fitness still needs to be calculated for the best program.
  unsigned int generations = std::min(generationsUsedOrNoSolution, c->maxGenerations);

  double bestTrainingFitness = best->Fitness();

  best->UpdateFitness(test);
  double bestTestFitness = best->Fitness();
  
  std::ofstream fout(c->runLogFilePath.c_str(), std::ios::out | std::ios::app);
  fout << generations << "," << time << "," << bestTrainingFitness << "," << bestTestFitness
       << std::endl;
  std::cout << generations << "," << time << "," << bestTrainingFitness << "," 
	    << bestTestFitness << std::endl;
  fout.close();

  delete pop;
  delete c;
  return 0;
}
